use galactica_content::{Content, EffectHandle, SpriteAutomaton};
use nalgebra::{Rotation2, Vector2};
use rand::Rng;
use rapier2d::dynamics::{RigidBodyBuilder, RigidBodyHandle, RigidBodyType};

use crate::phys::{PhysStepResources, PhysWrapper};

/// A single instance of an effect
#[derive(Debug, Clone)]
pub struct PhysEffect {
	/// The sprite to use for this effect
	pub anim: SpriteAutomaton,

	/// This effect's velocity, in world coordinates
	pub rigid_body: RigidBodyHandle,

	/// This effect's lifetime, in seconds
	lifetime: f32,

	/// The size of this effect,
	/// given as height in world units.
	pub size: f32,

	/// Fade this effect over this many seconds as it expires
	pub fade: f32,

	/// If true, this effect has been destroyed,
	/// and needs to be removed from the game.
	is_destroyed: bool,
}

impl PhysEffect {
	/// Create a new effect inside `Wrapper`
	pub fn new(
		ct: &Content,
		wrapper: &mut PhysWrapper,
		effect: EffectHandle,
		pos: Vector2<f32>,
		parent_angle: f32,
		parent_velocity: Vector2<f32>,
		target_velocity: Vector2<f32>,
	) -> Self {
		let effect = ct.get_effect(effect);
		let mut rng = rand::thread_rng();

		let velocity = {
			let a =
				rng.gen_range(-effect.velocity_scale_parent_rng..=effect.velocity_scale_parent_rng);
			let b =
				rng.gen_range(-effect.velocity_scale_target_rng..=effect.velocity_scale_target_rng);

			let velocity = ((effect.velocity_scale_parent + a) * parent_velocity)
				+ ((effect.velocity_scale_target + b) * target_velocity);

			Rotation2::new(rng.gen_range(-effect.direction_rng..=effect.direction_rng)) * velocity
		};

		let angvel = if effect.angvel_rng == 0.0 {
			effect.angvel
		} else {
			effect.angvel + rng.gen_range(-effect.angvel_rng..=effect.angvel_rng)
		};
		let angle = if effect.angle_rng == 0.0 {
			parent_angle + effect.angle
		} else {
			parent_angle + effect.angle + rng.gen_range(-effect.angle_rng..=effect.angle_rng)
		};

		let rb = RigidBodyBuilder::new(RigidBodyType::KinematicVelocityBased)
			.position(pos.into())
			.rotation(angle)
			.angvel(angvel)
			.linvel(velocity);

		PhysEffect {
			anim: SpriteAutomaton::new(ct, effect.sprite),
			rigid_body: wrapper.insert_rigid_body(rb.build()),
			lifetime: 0f32
				.max(effect.lifetime + rng.gen_range(-effect.lifetime_rng..=effect.lifetime_rng)),

			// Make sure size isn't negative. This check should be on EVERY rng!
			size: 0f32.max(effect.size + rng.gen_range(-effect.size_rng..=effect.size_rng)),
			fade: 0f32.max(effect.fade + rng.gen_range(-effect.fade_rng..=effect.fade_rng)),
			is_destroyed: false,
		}
	}

	/// Step this effect's state by `t` seconds
	pub fn step(&mut self, res: &PhysStepResources, wrapper: &mut PhysWrapper) {
		if self.is_destroyed {
			return;
		}

		self.anim.step(res.ct, res.t);
		self.lifetime -= res.t;

		if self.lifetime <= 0.0 {
			wrapper.remove_rigid_body(self.rigid_body);
			self.is_destroyed = true;
		}
	}

	/// Should this effect be deleted?
	pub fn is_destroyed(&self) -> bool {
		self.is_destroyed
	}

	/// The remaining lifetime of this effect, in seconds
	pub fn remaining_lifetime(&self) -> f32 {
		self.lifetime
	}
}
